//------------------------------------------------------------------------------
// Copyright (C) 2010 "Zalewa" <zalewapl@gmail.com>
//------------------------------------------------------------------------------
package zalewa.delayedproxy.network;

import java.io.IOException;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.util.LinkedList;
import java.util.Random;
import java.util.Vector;

import zalewa.delayedproxy.network.eventlisteners.PacketReceivedListener;
import zalewa.delayedproxy.network.eventlisteners.ThreadFatalErrorListener;

public class DatagramBase
{
	protected final static int PACKET_SIZE = 65536;

	
	
	private LinkedList<DelayedDatagramPacket> packetQueue = new LinkedList<DelayedDatagramPacket>();

	private int numPacketsLost = 0;
	private int numPacketsSent = 0;
	private int packetDelay = 0;
	private int packetLossChance = 0;
	private Vector<PacketReceivedListener> packetReceivedListenersArray = new Vector<PacketReceivedListener>();
	protected DatagramSocket socket = null;
	private Vector<ThreadFatalErrorListener> threadFatalErrorListenersArray = new Vector<ThreadFatalErrorListener>();	
	
	/**
	 * Thread associated with this DatagramBase (if any).
	 */
	public Thread thread = null;

	public void addPacketReceivedListener(PacketReceivedListener listener)
	{
		packetReceivedListenersArray.add(listener);
	}

	public synchronized void addPacketToQueue(DatagramPacket packet)
	{
		packetQueue.add(new DelayedDatagramPacket(packet, packetDelay));
	}
	
	public void addThreadFatalErrorListener(ThreadFatalErrorListener listener)
	{
		threadFatalErrorListenersArray.add(listener);
	}

	protected void cleanUpSocket()
	{
		try
		{
			socket.disconnect();
		}
		catch (Exception exc)
		{
			// Not an error.
		}

		try
		{
			socket.close();
		}
		catch (Exception exc)
		{
			// Not an error.
		}
	}

	protected void execPacketReceivedListeners(DatagramPacket packet)
	{
		for (PacketReceivedListener listener : this.packetReceivedListenersArray)
		{
			listener.packetReceived(this, packet);
		}
	}
	
	protected void execThreadFatalErrorListeners(Exception cause)
	{
		for (ThreadFatalErrorListener listener : this.threadFatalErrorListenersArray)
		{
			listener.fatalError(this, cause);
		}
	}

	public synchronized int getPacketDelayMs()
	{
		return packetDelay;
	}

	public synchronized int getPacketLossChance()
	{
		return packetLossChance;
	}

	/**
	 * This counts only the number of packets that were deliberately "lost" due
	 * to the packet loss chance. Packets lost through bad connection quality
	 * will not be accounted here.
	 * 
	 * @return Total number of packets lost by this instance.
	 */
	public synchronized int getNumPacketsLost()
	{
		return numPacketsLost;
	}

	/**
	 * @return Total number of packets SENT by this instance through
	 *         sendAllPackets() method.
	 */
	public synchronized int getNumPacketsSent()
	{
		return numPacketsSent;
	}

	protected boolean hasMorePackets()
	{
		return !packetQueue.isEmpty();
	}

	/**
	 * Seeks the first packet on the queue that is already after the delay.
	 * 
	 * @return Might return null if queue is empty or there are no delayed
	 *         packets yet.
	 */
	protected synchronized DatagramPacket popPacketFromQueue()
	{
		for (DelayedDatagramPacket delayedPacket : packetQueue)
		{
			if (delayedPacket.isAfterDelay())
			{
				packetQueue.remove(delayedPacket);

				Random random = new Random();
				int randomLossChance = random.nextInt(100);

				if (randomLossChance < packetLossChance)
				{
					++numPacketsLost;
					// Go to next packet.
					return popPacketFromQueue();
				}

				return delayedPacket.packet;
			}
		}

		return null;
	}

	/**
	 * Sends all packets without changing their address/port information.
	 */
	protected void sendAllPackets() throws IOException
	{
		DatagramPacket packet = popPacketFromQueue();
		while (packet != null)
		{
			sendPacket(packet, packet.getAddress(), packet.getPort());

			packet = popPacketFromQueue();
		}
	}
	
	protected void sendAllPackets(HostInfo targetInfo)
	throws IOException
	{
		sendAllPackets(targetInfo.address, targetInfo.port);
	}

	/**
	 * Sends all packets but changes their address/port information accordingly.
	 * 
	 * @param address
	 * @param port
	 * @throws IOException
	 */
	protected void sendAllPackets(InetAddress address, int port)
			throws IOException
	{
		DatagramPacket packet = popPacketFromQueue();
		while (packet != null)
		{
			sendPacket(packet, address, port);

			packet = popPacketFromQueue();
		}
	}

	private void sendPacket(DatagramPacket packet, InetAddress address, int port)
			throws IOException
	{
		packet.setAddress(address);
		packet.setPort(port);

		socket.send(packet);
		synchronized (this)
		{
			++numPacketsSent;
		}
	}

	public synchronized void setPacketDelayMs(int miliseconds)
	{
		if (miliseconds < 0)
		{
			miliseconds = 0;
		}

		packetDelay = miliseconds;
	}

	public synchronized void setPacketLossChance(int percentage)
	{
		if (percentage < 0)
		{
			percentage = 0;
		}

		if (percentage > 100)
		{
			percentage = 100;
		}

		packetLossChance = percentage;
	}
}
