//------------------------------------------------------------------------------
// Copyright (C) 2010 "Zalewa" <zalewapl@gmail.com>
//------------------------------------------------------------------------------
package zalewa.delayedproxy.network;

import java.net.DatagramPacket;
import java.net.MalformedURLException;
import java.net.UnknownHostException;
import java.util.Timer;
import java.util.TimerTask;
import java.util.Vector;

import zalewa.delayedproxy.exceptions.IncorrectPortNumberException;
import zalewa.delayedproxy.exceptions.LoopbackException;
import zalewa.delayedproxy.network.eventlisteners.PacketReceivedListener;
import zalewa.delayedproxy.network.eventlisteners.ThreadFatalErrorListener;

/**
 * 
 * The Proxy Manager, or rather the Proxy itself.<br>
 * <br>
 * This class manages two threads responsible for receiving and forwarding
 * packets. One thread receives packets from the client and forwards them to
 * another thread. This other thread sends the packets to the target host. At
 * the same time the other thread receives response packets from the server and
 * forwards them back to the first thread which then sends them back to the
 * client.<br>
 * <br>
 * Currently only single connection from the client is supported. If more
 * clients try to connect on the same port the results are unpredictable.
 * 
 */
public class ProxyManager
{
	class ClientPacketReceivedListener implements PacketReceivedListener
	{
		@Override
		public void packetReceived(DatagramBase packetReceiver, DatagramPacket packet)
		{
			HostInfo clientInfo = new HostInfo(packet.getAddress(), packet.getPort());

			ServerSender serverSender = findServerSender(clientInfo);
			serverSender.addPacketToQueue(packet);
		}
	}

	class ServerPacketReceivedListener implements PacketReceivedListener
	{
		@Override
		public void packetReceived(DatagramBase packetReceiver, DatagramPacket packet)
		{
			ServerSender sender = (ServerSender) packetReceiver;

			HostInfo clientInfo = sender.getClientInfo();

			packet.setAddress(clientInfo.address);
			packet.setPort(clientInfo.port);

			clientReceiver.addPacketToQueue(packet);
		}
	}

	private ClientReceiver clientReceiver = null;
	private Timer cleanUpTimer = new Timer();
	private int packetDelayMs = 0;
	private int packetLossPercentage = 0;
	private HostInfo targetHostInfo = null;
	private ThreadFatalErrorListener threadErrorListener = null;
	private Vector<ServerSender> serverSenders = new Vector<ServerSender>();

	/**
	 * Finds ServerSender that is responsible for handling specified connection
	 * from the client side. If ServerSender cannot be found this method will
	 * create one and append it to the end of serverSenders array.
	 * 
	 * @param clientInfo
	 * @return
	 */
	private ServerSender findServerSender(HostInfo clientInfo)
	{
		for (ServerSender sender : serverSenders)
		{
			if (sender.isClientInfoEqual(clientInfo))
			{
				return sender;
			}
		}

		return startServerSenderThread(clientInfo, this.targetHostInfo);
	}

	private void killDeadConnectionThreads()
	{
		for (int i = 0; i < serverSenders.size(); ++i)
		{
			ServerSender sender = serverSenders.elementAt(i);
			if (sender.isConnectionDead())
			{
				sender.quit();
				serverSenders.remove(i);
				--i;
			}
		}
	}

	public int getPacketDelay()
	{
		return this.packetDelayMs;
	}

	public int getPacketLossPercentage()
	{
		return packetLossPercentage;
	}

	public int getNumPacketsSentToClient()
	{
		return clientReceiver.getNumPacketsSent();
	}

	public int getNumPacketsLostToClient()
	{
		return clientReceiver.getNumPacketsLost();
	}

	public int getNumPacketsSentToServer()
	{
		int total = 0;
		for (ServerSender sender : serverSenders)
		{
			total += sender.getNumPacketsSent();
		}

		return total;
	}

	public int getNumPacketsLostToServer()
	{
		int total = 0;
		for (ServerSender sender : serverSenders)
		{
			total += sender.getNumPacketsLost();
		}

		return total;
	}

	public boolean isProxyRunning()
	{
		boolean bClientThread = isThreadRunning(clientReceiver);
		boolean bServerThread = isThreadRunning(serverSenders);

		return bClientThread || bServerThread;
	}

	private boolean isThreadRunning(DatagramBase datagramBase)
	{
		if (datagramBase == null)
		{
			return false;
		}
		
		return isThreadRunning(datagramBase.thread);
	}

	private boolean isThreadRunning(Vector<ServerSender> serverSendersArray)
	{
		for (ServerSender serverSender : serverSendersArray)
		{
			if (isThreadRunning(serverSender))
			{
				return true;
			}
		}

		return false;
	}

	private boolean isThreadRunning(Thread thread)
	{
		return thread != null && thread.isAlive();
	}

	/**
	 * Packet delay (or ping) is split half-by-half for ClientReceiver and
	 * ServerSender objects.<br>
	 * <br>
	 * For example if ping of 200 ms is specified receiver will wait 100 ms
	 * before forwarding a server's packet to the client and sender will wait
	 * 100 ms before sending a client's packet to the server.
	 * 
	 * @param miliseconds
	 *            A positive value or zero.
	 */
	public void setPacketDelay(int miliseconds)
	{
		if (miliseconds < 0)
		{
			miliseconds = 0;
		}

		this.packetDelayMs = miliseconds;

		if (clientReceiver != null)
		{
			clientReceiver.setPacketDelayMs(miliseconds / 2);
		}

		for (ServerSender serverSender : serverSenders)
		{
			serverSender.setPacketDelayMs(miliseconds / 2);
		}
	}

	/**
	 * Packet loss percentage is set equally to ClientReceiver and ServerSender
	 * objects.<br>
	 * <br>
	 * For example if a loss percentage of 50% is specified both receiver and
	 * server will have 50% probability of losing a packet.
	 * 
	 * @see setPacketDelay()
	 * @param percentage
	 *            A value between 0 and 100 inclusive.
	 */
	public void setPacketLossPercentage(int percentage)
	{
		if (percentage < 0)
		{
			percentage = 0;
		}

		if (percentage > 100)
		{
			percentage = 100;
		}

		this.packetLossPercentage = percentage;

		if (clientReceiver != null)
		{
			clientReceiver.setPacketLossChance(percentage);
		}

		for (ServerSender serverSender : serverSenders)
		{
			serverSender.setPacketLossChance(percentage);
		}
	}
	
	public void setThreadFatalErrorListener(ThreadFatalErrorListener listener)
	{
		this.threadErrorListener = listener;
	}

	private void startClientReceiverThread(int localPort)
	{
		clientReceiver = new ClientReceiver(localPort);
		
		if (this.threadErrorListener != null)
		{
			clientReceiver.addThreadFatalErrorListener(this.threadErrorListener);
		}
		
		clientReceiver.addPacketReceivedListener(new ClientPacketReceivedListener());
		clientReceiver.setPacketDelayMs(this.packetDelayMs / 2);
		clientReceiver.setPacketLossChance(this.packetLossPercentage);

		Thread clientReceiverThread = new Thread(clientReceiver);
		clientReceiverThread.start();

		clientReceiver.thread = clientReceiverThread;
	}

	private ServerSender startServerSenderThread(HostInfo clientInfo, HostInfo targetInfo)
	{
		ServerSender serverSender = new ServerSender(clientInfo, targetInfo);
		
		if (this.threadErrorListener != null)
		{
			serverSender.addThreadFatalErrorListener(this.threadErrorListener);
		}
		
		serverSender.addPacketReceivedListener(new ServerPacketReceivedListener());
		serverSender.setPacketDelayMs(this.packetDelayMs / 2);
		serverSender.setPacketLossChance(this.packetLossPercentage);

		Thread serverSenderThread = new Thread(serverSender);
		serverSenderThread.start();

		serverSender.thread = serverSenderThread;

		serverSenders.add(serverSender);

		return serverSender;
	}

	/**
	 * Starts the Proxy.
	 * 
	 * @param targetAddress
	 *            Target host address in format &lt;host&gt;[:&lt;port&gt;]
	 * @param localPort
	 *            Port on which the Proxy will be opened.
	 * @throws IncorrectPortNumberException
	 *             Thrown if port either in the targetAddress String or
	 *             localPort is incorrect.
	 * @throws MalformedURLException
	 *             Thrown if targetAddress is in incorrect format.
	 * @throws UnknownHostException
	 *             Thrown if targetAdress' host cannot be resolved.
	 */
	public void startThreads(String targetAddress, int localPort)
			throws LoopbackException, IncorrectPortNumberException, MalformedURLException,
			UnknownHostException
	{
		HostInfo.validatePort(localPort);
		targetHostInfo = new HostInfo(targetAddress);
		
		if (targetHostInfo.port == localPort)
		{
			throw new LoopbackException();
		}
		
		TimerTask timerTask = new TimerTask()
		{
			@Override
			public void run()
			{
				killDeadConnectionThreads();
			}
		};

		startClientReceiverThread(localPort);

		cleanUpTimer = new Timer();
		cleanUpTimer.scheduleAtFixedRate(timerTask,
				ServerSender.MAX_LAST_RECEIVED_TIME_MS,
				ServerSender.MAX_LAST_RECEIVED_TIME_MS);
	}

	/**
	 * Stops the Proxy. This method will block the calling thread until Proxy
	 * threads are finished.
	 */
	public void stopThreads()
	{
		cleanUpTimer.cancel();

		// Issue kill signals to threads.
		if (clientReceiver != null)
		{
			clientReceiver.quit();
		}

		for (ServerSender sender : serverSenders)
		{
			sender.quit();
		}

		// Wait till all threads are killed.
		try
		{
			while (isProxyRunning())
			{
				Thread.sleep(1);
			}
		}
		catch (InterruptedException e)
		{
			e.printStackTrace();
		}

		// Clean-up
		clientReceiver = null;
		serverSenders.clear();
	}
}
